/*
 * Copyright (C) 1999-2001 The Regents of the University of California
 * (through E.O. Lawrence Berkeley National Laboratory), subject to
 * approval by the U.S. Department of Energy.
 *
 * Use of this software is under license. The license agreement is included
 * in the file MVICH_LICENSE.TXT.
 *
 * Developed at Berkeley Lab as part of MVICH.
 *
 * Authors: Bill Saphir      <wcsaphir@lbl.gov>
 *          Michael Welcome  <mlwelcome@lbl.gov>
 */


/* Copyright (c) 2002-2005, The Ohio State University. All rights
 * reserved.
 *
 * This file is part of the MVAPICH software package developed by the
 * team members of The Ohio State University's Network-Based Computing
 * Laboratory (NBCL), headed by Professor Dhabaleswar K. (DK) Panda.
 *
 * For detailed copyright and licencing information, please refer to the
 * copyright file COPYRIGHT_MVAPICH in the top level MPICH directory.
 *
 */



#if defined(_SMP_) || defined(USE_MPD_BASIC) || defined(USE_MPD_RING)
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#endif

#ifdef MAC_OSX
#include <dlfcn.h>
#else
#include <sys/socket.h>
#endif
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include "viadev.h"
#include "viaparam.h"
#include "viapriv.h"
#include "vbuf.h"
#include "dreg.h"
#include "process/pmgr_client.h"
#include "mpid_smpi.h"

#define NIC_0 0
#define SUBCHANNEL_0 0

/* Internal functions used by IBAL */
#ifdef MCST_SUPPORT
#include <complib/cl_debug.h>
#include "bcast_info.h"

static void ib_ud_qp_init(void);
static void ud_mcst_av_init(void);
static void alloc_ack_region(void);
static void alloc_bcast_buffers(void);
#if defined(OSM_VENDOR_INTF_TS) || defined(OSM_VENDOR_INTF_MTL)
void create_mcgrp(void);
int join_mcgrp(void);
void leave_mcgrp(void);
int mlid;
#endif
void viadev_post_ud_recv(vbuf *);
void ud_cleanup(void);
int ud_prepost_count = 0;
int ud_prepost_threshold = VIADEV_UD_PREPOST_THRESHOLD;
int period = PERIOD;
int is_mcast_enabled = 1;
#endif

#ifdef MAC_OSX
extern int mvapich_need_malloc_init;
extern void *vt_dylib;
extern void *(*vt_malloc_ptr)(size_t);
extern void (*vt_free_ptr)(void *);
#endif


#ifdef USE_MPD_RING
static void ib_vapi_bootstrap_ring(int, int, char *, char *, char *,
                                   VIP_MEM_HANDLE *);
static int bootstrap_barrier(void);
static void ib_vapi_enable_ring(int, int, char *);
#endif

static void ib_vapi_init(int ts_type);
static void ib_vapi_qp_enable(void);
static void ib_vapi_finalize(void);
static void ib_rank_lid_table(int *alladdrs, int len);

#ifdef USE_INLINE
int inline_size = 0;
#endif

#ifdef USE_MPD_RING
static void *barrier_buff;
VIP_MEM_HANDLE *barrier_mem_hndl;
#endif

#if defined(_SMP_) || defined(USE_MPD_BASIC) || defined(USE_MPD_RING)
static int get_host_id(char *myhostname, int hostname_len);
#endif

#if defined(LAZY_MEM_UNREGISTER) && !defined(MAC_OSX)

/* NOTE: Inclusion of malloc.h this makefile will cause redundant
 * declarations of malloc, free, etc.  If compiling under
 * gcc with -Wredundant-decls AND -Werror, the compilation
 * of this module will fail.  We have turned off
 * -Wredundant-decls for this module only.  If and when
 * this is fixed in glibc, the makefile should be repaired.
 */
#include <malloc.h>

#ifdef MALLOC_HOOK
/* Initialize our malloc hook */
void (*__malloc_initialize_hook) (void) = mvapich_init_malloc_hook;
#endif

#endif

#ifdef MCST_SUPPORT
bcast_info_t *bcast_info = &(viadev.bcast_info);
#endif

static EVAPI_async_handler_hndl_t async_handler[MAX_NUM_HCAS];

static void async_event_handler(
		VAPI_hca_hndl_t hca_hndl,
		VAPI_event_record_t *event_p,
		void* priv_data)
{
	switch(event_p->type){
		case VAPI_QP_PATH_MIGRATED:
		case VAPI_EEC_PATH_MIGRATED:
		case VAPI_QP_COMM_ESTABLISHED:
		case VAPI_EEC_COMM_ESTABLISHED:
		case VAPI_SEND_QUEUE_DRAINED:
		case VAPI_PORT_ACTIVE:
		{
				D_PRINT("Got an asynchronous event: %s\n",
						VAPI_event_record_sym(event_p->type));
				break;
		}
		case VAPI_CQ_ERROR:
		case VAPI_LOCAL_WQ_INV_REQUEST_ERROR:
		case VAPI_LOCAL_WQ_ACCESS_VIOL_ERROR:
		case VAPI_LOCAL_WQ_CATASTROPHIC_ERROR:
		case VAPI_PATH_MIG_REQ_ERROR:
		case VAPI_LOCAL_EEC_CATASTROPHIC_ERROR:
		case VAPI_LOCAL_CATASTROPHIC_ERROR:
		case VAPI_PORT_ERROR:
		{
       	 	vapidev_error_abort(GEN_EXIT_ERR,
            	    "Got an asynchronous event: %s (%s)",
                	VAPI_event_record_sym(event_p->type),
                	VAPI_event_syndrome_sym(event_p->
                            syndrome));

				break;
		}
		default:
		    vapidev_error_abort(GEN_EXIT_ERR,
	                "Warning!! Got an undefined asynchronous event\n");

	}
}

int MPID_VIA_Init(int *argc, char ***argv, int *size, int *rank) 
{
    int i, j;
	int local_addr_len;
    int *alladdrs = NULL;
    int *local_addr = NULL;
    VAPI_ret_t ret;

#if (defined (_SMP_) && !((defined (USE_MPD_RING) || defined (USE_MPD_BASIC))))
    int hostid;
#endif

#ifdef MCST_SUPPORT
    char grp_info[GRP_INFO_LEN];

    if (NULL != getenv("DISABLE_HARDWARE_MCST")) {
        /* The user has disabled Multicast */
        is_mcast_enabled = 0;
    } else {
        /* The user has enabled Multicast */
        is_mcast_enabled = 1;
    }

#endif


#if defined(VIADEV_RPUT_SUPPORT) || defined(VIADEV_RGET_SUPPORT)
#ifdef LAZY_MEM_UNREGISTER
    /* Set glibc/stdlib malloc options to prevent handing
     * memory back to the system (brk) upon free.
     * Also, dont allow MMAP memory for large allocations.
     * This because we dont unregister memory when refcounts
     * go to zero.
     * Only need these options when RDMA operations are being
     * used.  Otherwise, MVICH layer owns all registered memory
     * and we will only free it if it is de-registered anyway.
     */
#ifdef MALLOC_HOOK
    create_hash_table();
#else
    mallopt(M_TRIM_THRESHOLD, -1);
    mallopt(M_MMAP_MAX, 0);
#endif 				/* MALLOC_HOOK */

#endif 				/* LAZY_MEM_UNREGISTER */
#endif		 		/* VIADEV_RPUT_SUPPORT */

    /* init the VIA barrier counter */
    viadev.barrier_id = 0;

    /*
     * Find out from the process manager who we are. 
     */
    if (!pmgr_client_init(argc,
   				          argv,
              			  &viadev.np, &viadev.me, &viadev.global_id)) {
	    vapidev_error_abort(GEN_EXIT_ERR, "pmgr_client_init failed");
    }

    *size = viadev.np;
    *rank = viadev.me;

   /* Now that we know our rank, we should also know
     * our name */
    viadev.my_name = (char *) malloc(sizeof(char) * HOSTNAME_LEN);

    if (NULL == viadev.my_name) {
        vapidev_error_abort(GEN_EXIT_ERR, "malloc failed");
    }

    if (gethostname(viadev.my_name, sizeof(char) * HOSTNAME_LEN) < 0) {
        vapidev_error_abort(GEN_EXIT_ERR, "gethostname failed");
    }


    D_PRINT("size=%d, myrank=%d\n", *size, *rank);

    SET_ORIGINAL_MALLOC_HOOKS;
   /* For both SMP and no-SMP, we have the same address length */
    alladdrs = (int *) malloc(2 * num_subchannels * viadev.np * sizeof(int) + viadev.np * sizeof(int));
    local_addr = (int *) malloc(2 * num_subchannels *  viadev.np * sizeof(int));

    viadev.lid_table = (int *) malloc(viadev.np * num_subchannels * sizeof(int));
    viadev.qp_table = (int *) malloc(viadev.np * num_subchannels * sizeof(int));

    if (alladdrs == NULL || local_addr == NULL ||
        viadev.lid_table == NULL || viadev.qp_table == NULL) {
        vapidev_error_abort(GEN_EXIT_ERR,
                            "malloc for alladdrs/local_addr/lid_table/qp_table");
    }


    SAVE_MALLOC_HOOKS;
    SET_MVAPICH_MALLOC_HOOKS;


    /* set run-time parameters that have been passed in through
     * environment variables
     */
    viadev_init_parameters(viadev.np, viadev.me);


#ifdef VIADEV_HAVE_RDMA_LIMIT
    viadev.outstanding_rdmas = 0;
#endif

    /* VAPI initializations */
    viadev.maxtransfersize = MAX_RDMA_SIZE;

    for (i = 0; i < num_hcas; i++) {
		ret = VAPI_open_hca(viadev_device[i], &(viadev.nic[i]));
		ret = EVAPI_get_hca_hndl(viadev_device[i], &(viadev.nic[i]));

	    if (VAPI_OK != ret) {
    		vapidev_error_abort(VAPI_RETURN_ERR, "cannot open HCA");
    	}

    }

    /* Get all LIDs by querying the HCA port */
    for (i = 0; i < num_subchannels; i++) {
		VAPI_hca_hndl_t hca = viadev.nic[i / num_ports];
    /* This is a case of looparound of ports */
	    int port = (viadev_default_port + i) % (num_ports + 1);
    	    if ( port < viadev_default_port)
        	    port += 1;


		ret = VAPI_query_hca_port_prop(hca,
			(IB_port_t)port, (VAPI_hca_port_t *)&(viadev.hca_port[i]));
    	if (VAPI_OK != ret) {
    		vapidev_error_abort(VAPI_RETURN_ERR, "cannot query  HCA for lid");
    	}
		
		viadev.my_hca_lid[i] = viadev.hca_port[i].lid;
		D_PRINT("lid=%d\n", viadev.my_hca_lid[i]);
    }

    /* Since we need to excahge qp # also, we will 
	 * create QP first, exchange addresses and 
     * then enable them
     */

    ib_vapi_init(IB_TS_RC);

    for (i = 0; i < num_hcas; i++) {
		ret = EVAPI_set_async_event_handler(viadev.nic[i], 
											async_event_handler, 0, 
											&(async_handler[i]));
    	if (VAPI_OK != ret) {
    		vapidev_error_abort(VAPI_RETURN_ERR,
            		    "Cannot set async_event_handler");
    	}

    }
#ifdef MCST_SUPPORT

    if (is_mcast_enabled) {

#if defined(MINISM)
        viadev.mcg_dgid[0] = 0xFF;
        viadev.mcg_dgid[1] = 0x00;
        viadev.mcg_dgid[2] = 0x00;
        viadev.mcg_dgid[3] = 0x00;
        viadev.mcg_dgid[4] = 0x00;
        viadev.mcg_dgid[5] = 0x00;
        viadev.mcg_dgid[6] = 0x00;
        viadev.mcg_dgid[7] = 0x00;
        viadev.mcg_dgid[8] = 0x00;
        viadev.mcg_dgid[9] = 0x00;
        viadev.mcg_dgid[10] = 0x00;
        viadev.mcg_dgid[11] = 0x00;
        viadev.mcg_dgid[12] = 0x00;
        viadev.mcg_dgid[13] = 0x00;
        viadev.mcg_dgid[14] = 0x00;
        viadev.mcg_dgid[15] = 0x00;
#endif

#if defined(OSM_VENDOR_INTF_TS) || defined(OSM_VENDOR_INTF_MTL)
        viadev.mcg_dgid[0] = 0xFF;
        viadev.mcg_dgid[1] = 0x12;
        viadev.mcg_dgid[2] = 0xa0;
        viadev.mcg_dgid[3] = 0x1c;
        viadev.mcg_dgid[4] = 0xfe;
        viadev.mcg_dgid[5] = 0x80;
        viadev.mcg_dgid[6] = 0x00;
        viadev.mcg_dgid[7] = 0x00;
        viadev.mcg_dgid[8] = 0x00;
        viadev.mcg_dgid[9] = 0x00;
        viadev.mcg_dgid[10] = 0x00;
        viadev.mcg_dgid[11] = 0x00;
        viadev.mcg_dgid[12] = 0x00;
        viadev.mcg_dgid[13] = 0x00;
        viadev.mcg_dgid[14] = 0x00;
        viadev.mcg_dgid[15] = 0x00;
#endif

        ib_ud_qp_init();        /* Allocate UD-QP */

#if defined(MINISM)
        ud_mcst_av_init();      /* Allocate address vect.for  multicast */
#endif
        alloc_ack_region();     /* Allocate Ack region for reliability  */

        alloc_bcast_buffers();  /* Allocate Bcast,Reliability structures  */
    }
#endif
#if defined(USE_MPD_RING) || defined (USE_MPD_BASIC)
    /*
     * For the scalable network discovery, we create a ring of queue pairs.
     * Then we exchange other queue_pair #'s over the ring.
     * In future, consider using MPD as a run_time option later.
     * For now, we stick with some of the process manager code.
     */

    {

        int temp_hostid;
        char *alladdrs_self;
        char *alladdrs_group;
        char *temp_ptr;

#ifdef USE_MPD_RING
        int lhs, rhs;
        VIP_MEM_HANDLE alladdrs_mr_hndl;
        char ring_qp_out[64];
        char ring_qp_in[128];

        temp_ptr = (char *) ring_qp_out;
        lhs = (viadev.me + viadev.np - 1) % viadev.np;
        rhs = (viadev.me + 1) % viadev.np;

        /* Prepare my_hca_lid, lhs qp, rsh qp for exchange */
        sprintf(temp_ptr, "%08d", viadev.my_hca_lid[0]);
        temp_ptr += 8;
        sprintf(temp_ptr, "%08d", viadev.boot_qp_prop[0].qp_num);
        temp_ptr += 8;
        sprintf(temp_ptr, "%08d", viadev.boot_qp_prop[1].qp_num);
        temp_ptr += 8;
        *temp_ptr = '\0';
        D_PRINT("ring_qp_out has %s\n", ring_qp_out);
#endif
        local_addr_len = (viadev.np * num_subchannels + 1) * QPLEN_XDR;
        alladdrs_self = (char *) malloc(local_addr_len + 4);
        alladdrs_group = (char *) malloc(viadev.np * local_addr_len + 4); 
	
        /* Prepare all the queue pairs from myself */
        temp_ptr = (char *) alladdrs_self;
	for (i = 0; i < viadev.np; ++i) {
	    for(j = 0; j < num_subchannels; j++) {
	       	if (i == viadev.me) {
		    sprintf(temp_ptr, "%08d", viadev.my_hca_lid[j]);
	       	} else {
		    sprintf(temp_ptr, "%08d", 
			    viadev.qp_prop[i * num_subchannels + j].qp_num);
	       	}
		temp_ptr += 8;
	    }
	}

        temp_hostid = get_host_id(viadev.my_name, HOSTNAME_LEN);
        sprintf(temp_ptr, "%08d", temp_hostid);
        temp_ptr += 8;
        *temp_ptr = '\0';

#ifdef USE_MPD_BASIC
        {
            int self_length = strlen(alladdrs_self);
            assert(self_length == local_addr_len);
            pmgr_exchange_addresses(alladdrs_self, self_length,
                                    alladdrs_group);
        }
#else
        /* Save the static buffer address */
        barrier_buff = (void *) alladdrs_group;
        barrier_mem_hndl = &alladdrs_mr_hndl;

        D_PRINT("Before exchange, out=%s\n", ring_qp_out);

        pmgr_exchange_addresses(ring_qp_out, 
		strlen(ring_qp_out), ring_qp_in);
        D_PRINT("After exchange, in=%s\n", ring_qp_in);

        /* Enable the ring to exhange all the others queue_pair numbers.
         * Consider rewriting all the following functions.
         */
        ib_vapi_bootstrap_ring(lhs, rhs, ring_qp_in, alladdrs_self,
                               alladdrs_group, &alladdrs_mr_hndl);
#endif
        D_PRINT("Done exchanging addresses group=%s\n", alladdrs_group);

        /*The memory allocated for alladdrs is not freed at this time. */
        ib_rank_lid_table((int *) alladdrs_group, local_addr_len);
        ib_vapi_qp_enable();

#ifdef USE_MPD_RING
        /* Free all the temorary memory used for MPD_RING */
        ret = VAPI_deregister_mr(viadev.nic[0], alladdrs_mr_hndl.hndl);
#endif

        free(alladdrs_self);
        free(alladdrs_group);
    }
#else   /* (defined(USE_MPD_BASIC) || defined (USE_MPD_RING)) */
    for (i = 0; i < viadev.np; ++i) {
		for(j = 0; j < num_subchannels; j++){
	        if (i == viadev.me) {
    	        local_addr[i * num_subchannels + j] = viadev.my_hca_lid[j];
        	} else {
            	local_addr[i* num_subchannels + j] = viadev.qp_prop[i * num_subchannels + j].qp_num;
        	}
    	}	
	}	
    local_addr_len = viadev.np * num_subchannels *sizeof(int);

#ifdef _SMP_
    hostid = get_host_id(viadev.my_name, HOSTNAME_LEN);
    local_addr[viadev.np * num_subchannels] = hostid;
    local_addr_len += sizeof(int);

#else /* put a useless integer for no-SMP mode */

    local_addr[viadev.np * num_subchannels] = viadev.me;
    local_addr_len += sizeof(int);
#endif
	
    pmgr_exchange_addresses(local_addr, local_addr_len, alladdrs);

    ib_rank_lid_table(alladdrs, local_addr_len);
    ib_vapi_qp_enable();
#endif  /* End of ! (defined(USE_MPD_BASIC) || defined (USE_MPD_RING)) */


#if defined(VIADEV_RPUT_SUPPORT)
    dreg_init(viadev.nic, viadev.ptag);
    D_PRINT("dreg_init done\n");
#endif

#if defined(MCST_SUPPORT) && (defined(OSM_VENDOR_INTF_TS) || \
    defined(OSM_VENDOR_INTF_MTL))

    if (is_mcast_enabled) {
        if (viadev.me == 0) {
            create_mcgrp();     /* Creation of mcst group */
        }

        if (viadev.me == 0) {
            strncpy(grp_info, "MCS_Grp1", GRP_INFO_LEN);
        }
    }

    pmgr_exchange_mcs_group_sync(grp_info, GRP_INFO_LEN, 0);

    if (is_mcast_enabled) {
        mlid = join_mcgrp();           /* Joining the group */
        ud_mcst_av_init();      /* Allocate address vect. for multicast */

        /* Preposting the descriptors for the UD transport */
        for (i = 0; i < VIADEV_UD_PREPOST_DEPTH; i++) {
            PREPOST_UD_VBUF_RECV();
            ud_prepost_count++;
        }
    }
#endif
#ifdef MAC_OSX
    mvapich_malloc_init();
    mvapich_need_malloc_init = 0;
#endif
    return VAPI_OK;

}

/*
 * Set up all communication and put VIs in the VI array in viadev
 * The algorithm is as follows (may change when M-VIA supports peer-to-peer
 * connections):
 *   . use client-server routines to establish connections
 *   . for each pair, the lower-numbered process acts as the server
 *   . we should be careful not to establish hotspots. E.g., if
 *     every process connects with others in order 0, 1, 2. We
 *     should also be careful not to serialize.
 *     there is a nice fancy way to do this, with a communication pattern
 *     essentially the same as in an FFT.
 *   . However, it is a bit tricky to get right so for now
 *     do the stupid serialized version (xxx):
 *        for (i = 0; i < me; i++) connect(i)
 *        for (i = me+1; i <= np; i++) accept(i)
 *
 *  Note that we do not set up a connection to ourselves
 *
 */

void MPID_Wtime(double *t)
{
    struct timeval tv;
    static int initialized = 0;
    static int sec_base;
    gettimeofday(&tv, NULL);
    if (!initialized) {
    sec_base = tv.tv_sec;
    initialized = 1;
    }
    *t = (double) (tv.tv_sec - sec_base) + (double) tv.tv_usec * 1.0e-6;
}


void MPID_Wtick(double *t)
{
  *t = VIADEV_WTICK;
}


void MPID_VIA_Finalize(void)
{
	int ret;
	int i;

	D_PRINT("entering MPID_VIA_Finalize ...\n");


	for (i = 0; i < num_hcas; i++) {
		ret = EVAPI_clear_async_event_handler(viadev.nic[i], async_handler[i]);
    	if (VAPI_OK != ret) {
    		vapidev_error_abort(VAPI_RETURN_ERR,
            	    "Cannot ear async_event_handler");
    	}

	}

	/* VAPI finalize called */
	ib_vapi_finalize();

	/* free connection array space */
	if (viadev.connections) 
		free(viadev.connections);


	/*	let the process manager reclaim data strucrures 
	*	before termination */
	pmgr_finalize();

    if (viadev.lid_table) {
    	free(viadev.lid_table);
    }
    if (viadev.qp_table) {
    	free(viadev.qp_table);
    }
#ifdef _SMP_
    if (smpi.local_nodes) {
        free(smpi.local_nodes);
    }
#endif

#ifdef MCST_SUPPORT
        /* complib_exit() added for fixing comlib issue on RH-AS3,RH90,SLES9.
                  Works only with osm-mvapi-0.3.2-rc1 */
        complib_exit();
#endif
        

	D_PRINT("exiting MPID_VIA_Finalize...\n");

}
/* internal functions used for initialization and finalization by VAPI */

static void ib_vapi_init(int ts_type)
{

    VAPI_qp_init_attr_t qp_query_init_attr, qp_init_attr;
    VAPI_qp_attr_t qp_query_attr;
    VAPI_qp_attr_mask_t qp_query_attr_mask;

    call_result_t ret;
    unsigned int act_num_cqe;
    int i, j;

    /* allocate completion queues, protection handle before qp */
    for (i = 0; i < num_hcas; i++) {
		ret = VAPI_alloc_pd(viadev.nic[i], &(viadev.ptag[i]));
    	if (VAPI_OK != ret) {
	    	vapidev_error_abort(VAPI_RETURN_ERR, "Cannot allocate PD");
    	}

		ret = VAPI_create_cq(viadev.nic[i], viadev_cq_size, 
		 	  				&(viadev.cq_hndl[i]), &act_num_cqe);
	    if (VAPI_OK != ret) {
    		vapidev_error_abort(VAPI_RETURN_ERR, "Cannot allocate CQ");
    	}
		D_PRINT("Created CQ = %d with %d entries\n",viadev.cq_hndl[i], act_num_cqe);
    }

    SET_ORIGINAL_MALLOC_HOOKS;

#ifdef USE_MPD_RING
    ret = VAPI_create_cq(viadev.nic[0], viadev_cq_size,
                         &viadev.boot_cq_hndl, &act_num_cqe);

    if (VAPI_OK != ret) {
        vapidev_error_abort(VAPI_RETURN_ERR, "cannot allocate CQ\n");
    }
#endif

    viadev.qp_hndl = (VAPI_qp_hndl_t *)
	malloc(sizeof(VAPI_qp_hndl_t) * viadev.np * num_subchannels);

    if (NULL == viadev.qp_hndl) {
   		vapidev_error_abort(GEN_EXIT_ERR,
                "Cannot allocate memory for QPs\n");
    }
    
    viadev.qp_prop = (VAPI_qp_prop_t *)
				  	 malloc(sizeof(VAPI_qp_prop_t) * viadev.np * num_subchannels);

    if (NULL == viadev.qp_prop) {
		vapidev_error_abort(GEN_EXIT_ERR,
                "Cannot allocate memory for QP properties\n");
    }

    SAVE_MALLOC_HOOKS;
    SET_MVAPICH_MALLOC_HOOKS;

    qp_init_attr.cap.max_oust_wr_rq = viadev_rq_size;
    qp_init_attr.cap.max_oust_wr_sq = viadev_sq_size;
    qp_init_attr.cap.max_sg_size_rq = viadev_default_max_sg_list;
    qp_init_attr.cap.max_sg_size_sq = viadev_default_max_sg_list;
    qp_init_attr.rdd_hndl           = 0;

    qp_init_attr.rq_sig_type = VAPI_SIGNAL_REQ_WR;
    qp_init_attr.sq_sig_type = VAPI_SIGNAL_REQ_WR;
    qp_init_attr.ts_type = ts_type;

    for (i = 0; i < viadev.np; i++) {

	    if (i == viadev.me) {
			continue;
		}

    for (j = 0; j < num_subchannels; j++) {
		int qp_index;
		VAPI_hca_hndl_t hca = viadev.nic[j / num_ports];
		qp_index = i * num_subchannels + j;

		qp_init_attr.pd_hndl = viadev.ptag[j/num_ports];
	    qp_init_attr.rq_cq_hndl = viadev.cq_hndl[j/num_ports];
		qp_init_attr.sq_cq_hndl = viadev.cq_hndl[j/num_ports];

	    ret= VAPI_create_qp(hca, &qp_init_attr, &viadev.qp_hndl[qp_index],
				   	        &viadev.qp_prop[qp_index]);

	    if (VAPI_OK != ret) {
    	    vapidev_error_abort(VAPI_RETURN_ERR, "Could not create QP");
    	}
   
	   	D_PRINT("QP[%d] created ...qpnum=%d..hndl=%d\n",qp_index, 
				viadev.qp_prop[qp_index].qp_num,
				viadev.qp_hndl[qp_index]);

#ifdef USE_INLINE
		QP_ATTR_MASK_CLR_ALL(qp_query_attr_mask);
		ret = VAPI_query_qp(hca, viadev.qp_hndl[qp_index], &qp_query_attr, 
			  &qp_query_attr_mask, &qp_query_init_attr);
	    if (ret != VAPI_OK) {
    	    vapidev_error_abort(VAPI_RETURN_ERR,
        	        "Error: Querying QP : %s\n",
            	    VAPI_strerror(ret));
	    }
	
  		inline_size = qp_query_attr.cap.max_inline_data_sq;
    
#endif
    } 
}
#ifdef USE_MPD_RING
    qp_init_attr.cap.max_oust_wr_rq = viadev_rq_size + viadev.np;
    qp_init_attr.rq_cq_hndl = viadev.boot_cq_hndl;
    qp_init_attr.sq_cq_hndl = viadev.boot_cq_hndl;

    ret = VAPI_create_qp(viadev.nic[0], &qp_init_attr,
                         &viadev.boot_qp_hndl[0], &viadev.boot_qp_prop[0]);
    if (VAPI_OK != ret) {
        vapidev_error_abort(VAPI_RETURN_ERR, "Could not create lhs QP");
    }

    ret = VAPI_create_qp(viadev.nic[0], &qp_init_attr,
                         &viadev.boot_qp_hndl[1], &viadev.boot_qp_prop[1]);
    if (VAPI_OK != ret) {
        vapidev_error_abort(VAPI_RETURN_ERR, "Could not create rhs QP");
    }
#endif


}

static void ib_vapi_qp_enable(void)
{
    VAPI_qp_attr_t qp_attr[MAX_SUBCHANNELS];
    VAPI_qp_cap_t qp_cap[MAX_SUBCHANNELS];
    VAPI_qp_attr_mask_t qp_attr_mask[MAX_SUBCHANNELS];

#if defined(RDMA_FAST_PATH)

    VIP_MEM_HANDLE mem_handle;
    VAPI_mrw_t mr_in, mr_out;
    VAPI_ret_t result;

#ifdef SINGLE_RDMA_BUFFER
    int RDMA_buf_size = 0;
    char *RDMA_recv_buf = NULL;
    char *RDMA_send_buf = NULL;
    int pagesize;
    int rank_cnt;
#endif
#endif
    int i, j, k;
    call_result_t ret;
   
    /* Modifying  QP to INIT */
	for (j = 0; j < num_subchannels; j++){
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask[j]);
    qp_attr[j].qp_state = VAPI_INIT;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_QP_STATE);
    qp_attr[j].pkey_ix  = viadev_default_pkey_ix;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_PKEY_IX);
    qp_attr[j].port = viadev_default_port;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_PORT);
    qp_attr[j].remote_atomic_flags = VAPI_EN_REM_WRITE | VAPI_EN_REM_READ;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_REMOTE_ATOMIC_FLAGS);
	}
    for (i = 0; i < viadev.np ; i++){

		if(i == viadev.me )	{	 
			continue;
		}

		for (j = 0; j < num_subchannels; j++) {
        qp_attr[j].port=  (viadev_default_port + j) % (num_ports + 1);
        if ( qp_attr[j].port < viadev_default_port )
            qp_attr[j].port += 1;

	    ret = VAPI_modify_qp(viadev.nic[j/num_ports], 
							viadev.qp_hndl[i * num_subchannels + j], 
							&qp_attr[j], &qp_attr_mask[j], &qp_cap[j]);

	    if (VAPI_OK != ret)
    	    vapidev_error_abort(VAPI_RETURN_ERR,
        	        "Could not modify qp to INIT");

	    D_PRINT("Modified to init..Qp[%d]\n",i * num_subchannels + j);
		}
    }

    /**********************  INIT --> RTR  ************************/
	for (j = 0; j < num_subchannels; j++) {
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask[j]);
    qp_attr[j].qp_state = VAPI_RTR;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_QP_STATE);
    qp_attr[j].qp_ous_rd_atom = viadev_default_qp_ous_rd_atom;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_QP_OUS_RD_ATOM);
    qp_attr[j].path_mtu = viadev_default_mtu;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_PATH_MTU);
    qp_attr[j].rq_psn = viadev_default_psn;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_RQ_PSN);
    qp_attr[j].pkey_ix = viadev_default_pkey_ix;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_PKEY_IX);
    qp_attr[j].min_rnr_timer = viadev_default_min_rnr_timer;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_MIN_RNR_TIMER);

    qp_attr[j].av.sl = viadev_default_service_level;
    qp_attr[j].av.grh_flag = FALSE;
    qp_attr[j].av.src_path_bits = viadev_default_src_path_bits;
    }

    for ( i = 0; i < viadev.np; i++){

	    if(i == viadev.me) {
			continue;
		}

	    for (j = 0; j < num_subchannels; j++) {
			qp_attr[j].av.static_rate = link_speed[j]; 
			qp_attr[j].dest_qp_num = rank_to_qp_num(i, j); 
			QP_ATTR_MASK_SET(qp_attr_mask[j], QP_ATTR_DEST_QP_NUM);
			qp_attr[j].av.dlid = rank_to_hca_lid(i, j);
			QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_AV);
			D_PRINT("Remote LID[%d] QP=%d\n",qp_attr[j].av.dlid,qp_attr[j].dest_qp_num);

			ret = VAPI_modify_qp(viadev.nic[j/num_ports], 
				  viadev.qp_hndl[i * num_subchannels + j], 
				  &qp_attr[j], &qp_attr_mask[j], &qp_cap[j]);

		    if (VAPI_OK != ret)
        		vapidev_error_abort(VAPI_RETURN_ERR,
                	"Could not modify qp to RTR");

			D_PRINT("Modified to RTR..Qp[%d]\n",i * num_subchannels + j);
    	}
    }


    /************** RTS *******************/
	for (j = 0; j < num_subchannels; j++) {
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask[j]);
    qp_attr[j].qp_state = VAPI_RTS;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_QP_STATE);
    qp_attr[j].sq_psn = viadev_default_psn;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_SQ_PSN);
    qp_attr[j].timeout = viadev_default_time_out;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_TIMEOUT);
    qp_attr[j].retry_count = viadev_default_retry_count; 
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_RETRY_COUNT);
    qp_attr[j].rnr_retry = viadev_default_rnr_retry;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_RNR_RETRY);
	qp_attr[j].ous_dst_rd_atom = viadev_default_max_rdma_dst_ops;
    QP_ATTR_MASK_SET(qp_attr_mask[j],QP_ATTR_OUS_DST_RD_ATOM);

	}
    for(i = 0 ; i < viadev.np; i++){

	    if(i == viadev.me )	{
			continue;
		}

	    for (j = 0; j < num_subchannels; j++) {

  		ret = VAPI_modify_qp(viadev.nic[j/num_ports], 
							 viadev.qp_hndl[i * num_subchannels + j], 
							 &qp_attr[j], &qp_attr_mask[j], &qp_cap[j]);

	    if (VAPI_OK != ret)
    	    vapidev_error_abort(VAPI_RETURN_ERR,
        	        "Could not modify qp to RTS");

	   	D_PRINT("Modified to RTS..Qp[%d]\n",i * num_subchannels + j);
    	}
    }

    /***********************prepost receive descriptors ****************/

	allocate_vbufs(viadev.nic, viadev.ptag, viadev_vbuf_pool_size);

    SET_ORIGINAL_MALLOC_HOOKS;

    /* initialize connection parameters */
    viadev.connections = (viadev_connection_t *)
						  malloc(sizeof(viadev_connection_t) * viadev.np);

    if (NULL == viadev.connections) {
    	vapidev_error_abort(GEN_EXIT_ERR,
        	        "Unable to malloc connection structures");
    }


#if defined(RDMA_FAST_PATH) && defined(SINGLE_RDMA_BUFFER)
    pagesize = getpagesize();
    RDMA_buf_size =
        2 * (sizeof(struct vbuf) * viadev_num_rdma_buffer *
             (viadev.np - 1) + 2 * pagesize);
    viadev.RDMA_bufs = (char *) malloc(RDMA_buf_size);

    if (NULL == viadev.RDMA_bufs) {
        vapidev_error_abort(GEN_EXIT_ERR,
                            "Unable to allocate Fast RDMA buffers");
    }
    memset(viadev.RDMA_bufs, 0, RDMA_buf_size);

    /* register RDMA buffers */
    mr_in.acl = VAPI_EN_LOCAL_WRITE | VAPI_EN_REMOTE_WRITE;
    mr_in.l_key = 0;
    mr_in.r_key = 0;
    mr_in.pd_hndl = viadev.ptag;
    mr_in.size = RDMA_buf_size;
    mr_in.start = (VAPI_virt_addr_t) (virt_addr_t) (viadev.RDMA_bufs);
    mr_in.type = VAPI_MR;

    result = VAPI_register_mr(viadev.nic, &mr_in,
                              &mem_handle.hndl, &mr_out);

    if (result != VAPI_OK) {
        vapidev_error_abort(VAPI_RETURN_ERR, "VAPI_register_mr");
    }

    mem_handle.lkey = mr_out.l_key;
    mem_handle.rkey = mr_out.r_key;
    viadev.RDMA_buf_hndl = mem_handle;

    /* page aligned buffers for RDMA send and recv */
    RDMA_send_buf =
        (char *) (((unsigned long) viadev.RDMA_bufs + pagesize - 1) /
                  pagesize * pagesize);
    RDMA_recv_buf =
        (char *) (RDMA_send_buf +
                  (RDMA_buf_size / 2 + pagesize -
                   1) / pagesize * pagesize);

    D_PRINT("RDMA_send_buf:%p RDMA_recv_buf:%p, size:%d\n", RDMA_send_buf,
            RDMA_recv_buf, RDMA_buf_size);

    rank_cnt = 0;
#endif                          /* RDMA_FAST_PATH && SINGLE_RDMA_BUFFER */

    SAVE_MALLOC_HOOKS;
    SET_MVAPICH_MALLOC_HOOKS;

    for (i = 0; i < viadev.np; i++) {

		viadev_connection_t *c = &viadev.connections[i];
		c->global_rank = i;
		c->next_packet_expected = 1;
		c->next_packet_tosend = 1;
		c->shandle_head = NULL;
		c->shandle_tail = NULL;
		c->rhandle      = NULL;
		c->nextflow     = NULL;
		c->inflow       = 0;
		c->rendezvous_packets_expected = 0;
		viadev_init_backlog_queue(c);

    	for (j = 0; j < num_subchannels; j++) {

#ifdef RDMA_FAST_PATH
        	VAPI_hca_hndl_t hca = viadev.nic[j/num_ports];
#endif
	        c->preposts[j]     = 0;

#ifdef RDMA_FAST_PATH
#ifndef DISABLE_HEADER_CACHING
    /* init cached packet header */
        memset(&(c->cached_outgoing), 0,
               sizeof(viadev_packet_eager_start));
        memset(&(c->cached_incoming), 0,
               sizeof(viadev_packet_eager_start));
        c->cached_hit = c->cached_miss = 0;
        c->cached_incoming.header.type
        = c->cached_outgoing.header.type
        = FAST_EAGER_CACHED;
#endif
#endif
     	   if (i != viadev.me) {
        		c->vi[j] = viadev.qp_hndl[i * num_subchannels + j];

                for (k = 0; k < viadev_initial_prepost_depth; k++) {
                    PREPOST_VBUF_RECV(c, j);
                }
                D_PRINT("Preposted %d receive desc\n",viadev_prepost_depth);

                c->send_wqes_avail[j] = viadev_sq_size;
                c->ext_sendq_head[j] = NULL;
                c->ext_sendq_tail[j] = NULL;

                c->remote_credit[j] = viadev_initial_credits;
                c->remote_cc[j] = viadev_initial_credits;
                c->local_credit[j] = 0;
                c->preposts[j] = viadev_initial_credits;
               c->initialized[j] = (viadev_initial_prepost_depth
                              ==
                              viadev_prepost_depth +
                              viadev_prepost_noop_extra);
        	} else {
	        /* these should be ignored for self-communication */

		        c->vi[j] = VAPI_INVAL_HNDL;
    		    c->remote_credit[j] = 0;
        		c->remote_cc[j] = 0;
        		c->local_credit[j] = 0;
        	}

#if defined(RDMA_FAST_PATH)

        if (i != viadev.me) {

	        /* initialize revelant fields */
    	    c->rdma_credit[j] = 0;
        	c->num_no_completion[j] = 0;

			if (viadev_num_rdma_buffer) {
#ifdef SINGLE_RDMA_BUFFER
                c->RDMA_send_buf =
                    (struct vbuf *) ((unsigned long) RDMA_send_buf +
                                     rank_cnt * sizeof(struct vbuf) *
                                     viadev_num_rdma_buffer);
                c->RDMA_recv_buf =
                    (struct vbuf *) ((unsigned long) RDMA_recv_buf +
                                     rank_cnt * sizeof(struct vbuf) *
                                     viadev_num_rdma_buffer);
                rank_cnt++;
#else


	        SET_ORIGINAL_MALLOC_HOOKS;

#define RDMA_ALIGNMENT 64
#define RDMA_ALIGNMENT_OFFSET (64)
        /* allocate RDMA buffers */
    	    c->RDMA_send_buf_orig[j] =
        	    malloc(sizeof(struct vbuf) * viadev_num_rdma_buffer +
            	   2 * RDMA_ALIGNMENT);
	        c->RDMA_recv_buf_orig[j] =
    	        malloc(sizeof(struct vbuf) * viadev_num_rdma_buffer +
    	           2 * RDMA_ALIGNMENT);
	
        /* align vbuf->buffer to 64 byte boundary */
        	c->RDMA_send_buf[j] =
            	(struct vbuf
	             *) ((unsigned long) (c->RDMA_send_buf_orig[j])
    	         / RDMA_ALIGNMENT * RDMA_ALIGNMENT +
        	     RDMA_ALIGNMENT + RDMA_ALIGNMENT_OFFSET);
	        c->RDMA_recv_buf[j] =
    	        (struct vbuf
        	     *) ((unsigned long) (c->RDMA_recv_buf_orig[j])
            	 / RDMA_ALIGNMENT * RDMA_ALIGNMENT +
	             RDMA_ALIGNMENT + RDMA_ALIGNMENT_OFFSET);

        /* zero buffers */
    	    memset(c->RDMA_send_buf[j], 0,
        	       sizeof(struct vbuf) * viadev_num_rdma_buffer);
	        memset(c->RDMA_recv_buf[j], 0,
    	           sizeof(struct vbuf) * viadev_num_rdma_buffer);


        	SAVE_MALLOC_HOOKS;
	        SET_MVAPICH_MALLOC_HOOKS;
#endif
        /* set pointers */
    	    c->phead_RDMA_send[j] = 0;
        	c->ptail_RDMA_send[j] = viadev_num_rdma_buffer - 1;

	        c->p_RDMA_recv[j] = 0;
    	    c->p_RDMA_recv_tail[j] = viadev_num_rdma_buffer - 1;

#ifndef SINGLE_RDMA_BUFFER
    	    /* register RDMA buffers */
	        mr_in.acl = VAPI_EN_LOCAL_WRITE|VAPI_EN_REMOTE_WRITE;
	        mr_in.l_key = 0;
    	    mr_in.r_key = 0;
	        mr_in.pd_hndl = viadev.ptag[j/num_ports];
    	    mr_in.size = sizeof(struct vbuf) * viadev_num_rdma_buffer;
        	mr_in.start = (VAPI_virt_addr_t)(virt_addr_t)(c->RDMA_send_buf[j]);
	        mr_in.type = VAPI_MR;

    	    result = VAPI_register_mr(hca, &mr_in, &mem_handle.hndl, &mr_out);
        	if (result != VAPI_OK) {
                    vapidev_error_abort(VAPI_RETURN_ERR,
                                        "VAPI_register_mr");
        	}

	        mem_handle.lkey = mr_out.l_key;
    	    mem_handle.rkey = mr_out.r_key;
#endif

        	c->RDMA_send_buf_hndl[j] = mem_handle;

#ifndef SINGLE_RDMA_BUFFER
	        mr_in.acl = VAPI_EN_LOCAL_WRITE|VAPI_EN_REMOTE_WRITE;
    	    mr_in.l_key = 0;
	        mr_in.r_key = 0;
    	    mr_in.pd_hndl = viadev.ptag[j/num_ports];
       		mr_in.size = sizeof(struct vbuf) * viadev_num_rdma_buffer;
	        mr_in.start = (VAPI_virt_addr_t)(virt_addr_t)(c->RDMA_recv_buf[j]);
    	    mr_in.type = VAPI_MR;

        	result = VAPI_register_mr(hca, &mr_in, &mem_handle.hndl, &mr_out);
	        mem_handle.lkey = mr_out.l_key;
    	    mem_handle.rkey = mr_out.r_key;

        	c->RDMA_recv_buf_hndl[j] = mem_handle;
#endif
            } else {
                c->phead_RDMA_send[j] = 0;
                c->ptail_RDMA_send[j] = 0;
                c->p_RDMA_recv[j] = 0;
                c->p_RDMA_recv_tail[j] = 0;
            }
		
        	/* remote address and handle not available yet */
	        c->remote_address_received = 0;

    	    } /* if i != viadev.me */
#endif
    	} 	/* for each subchannels */
    } 		/* for each connection */


#if defined(MCST_SUPPORT) && defined(MINISM)
    if (is_mcast_enabled) {
        /* Preposting the descriptors for the UD transport */
        for (i = 0; i < VIADEV_UD_PREPOST_DEPTH; i++) {
            PREPOST_UD_VBUF_RECV();
            ud_prepost_count++;
        }
    }
#endif                          /* MCST_SUPPORT && MINISM */

    /* call barrier to make sure all preposting has been done */

    D_PRINT("about to call pmgr_barrier...\n");

#ifdef USE_MPD_RING
    i = bootstrap_barrier();
#else
    i = pmgr_barrier();
#endif                          /* USE_MPD_RING */

    D_PRINT("process manager barrier returns %d\n", i);

    if (viadev_num_rdma_buffer == 0)
        return;

    	/* for out-of-order message handling */
#ifdef RDMA_FAST_PATH
	    init_ofo_queues();
#endif
    	D_PRINT("ofo queues initiated\n");

    /* set up information for weighted striping based on link speeds */
    for (i = 0; i < viadev.np; i++) {
	    viadev_connection_t *c = &viadev.connections[i];
    	for (j = 0; j < num_subchannels; j++) {
				c->s_weight[j] = 4; 
			/* 4X speed */
    	}
    }


#if defined(RDMA_FAST_PATH) || defined(MCST_SUPPORT)
    /* inform others about the receive 
		RDMA buffer address and handle */

    for(i = 0; i < viadev.np; i++)
    {

	    if (i == viadev.me) {
			continue;
		}

    	for (j = 0; j < num_subchannels; j++) {
    		int subchannel;
		    vbuf *v;
		    viadev_packet_rdma_address *h;
		    viadev_connection_t *c = &viadev.connections[i];

		    v = get_vbuf();
		    h = (viadev_packet_rdma_address *)VBUF_BUFFER_START(v);

        	/* set up the packet */
		    PACKET_SET_HEADER(h, c, VIADEV_RDMA_ADDRESS);
		    h->envelope.src_lrank = i;

		    /* set address and handle */
#if defined(RDMA_FAST_PATH)
		    h->RDMA_address = c->RDMA_recv_buf[j];
		    h->RDMA_hndl = c->RDMA_recv_buf_hndl[j];
#endif
		    h->subchannel = j;

		    /* finally send it off */
		    /* always use the one channel to avoid out of order issue unless 
			 * doing some fancy stuff for scheduling
	 		*/

#ifdef MCST_SUPPORT
        if (is_mcast_enabled) {
            /* set address and handle for UD Acks, this is for now remove later */
            h->RDMA_ACK_address = (viadev.bcast_info.ack_buffer);
            h->RDMA_ACK_hndl.rkey = viadev.bcast_info.ack_mem_hndl.rkey;
        }
#endif                          /* MCST_SUPPORT */

		    subchannel = choose_sr_subchannel(c);
		    vbuf_init_send(v, sizeof(viadev_packet_rdma_address), subchannel);
		    viadev_post_send(c, v, subchannel);
    	}
    }
    /* get all the address */
    for(i = 0; i < viadev.np; i++)
    {
	    viadev_connection_t *c = &viadev.connections[i];

    	if (i == viadev.me) {
			continue;
		}
    	while (c->remote_address_received < num_subchannels) {
			MPID_DeviceCheck(1);
		}
    }

#endif				/* RDMA_FAST_PATH || MCST_SUPPORT */
}



#if defined(RDMA_FAST_PATH)
void process_remote_rdma_address(	viadev_connection_t * c,
                                 	VIP_MEM_HANDLE remote_hndl,
                                 	struct vbuf *remote_address,
									int subchannel);

void process_remote_rdma_address(   viadev_connection_t * c,
                                    VIP_MEM_HANDLE remote_hndl,
                                    struct vbuf *remote_address,
                                    int subchannel)
{
    int i;

    c->remote_address_received++;
    c->remote_RDMA_buf_hndl[subchannel] 
			= remote_hndl;
    c->remote_RDMA_buf[subchannel] 
			= remote_address;

    for (i = 0; i < viadev_num_rdma_buffer; i++) {

		/* set global rank */
		c->RDMA_send_buf[subchannel][i].grank = c->global_rank;
		c->RDMA_recv_buf[subchannel][i].grank = c->global_rank;
		c->RDMA_send_buf[subchannel][i].subchannel = subchannel;
		c->RDMA_recv_buf[subchannel][i].subchannel = subchannel;

		c->RDMA_recv_buf[subchannel][i].desc.next = NULL;

		/* fill out the descriptors */
		c->RDMA_send_buf[subchannel][i].padding = FREE_FLAG;
		
		c->RDMA_send_buf[subchannel][i].desc.sg_entry.len = 0;

		c->RDMA_send_buf[subchannel][i].desc.sg_entry.addr = (VAPI_virt_addr_t)
					(MT_virt_addr_t)(&((c->RDMA_send_buf[subchannel] + i)->head_flag));
		c->RDMA_send_buf[subchannel][i].desc.sg_entry.lkey = 
					c->RDMA_send_buf_hndl[subchannel].lkey;
	
		c->RDMA_send_buf[subchannel][i].desc.sr.sg_lst_p = 
					&(c->RDMA_send_buf[subchannel][i].desc.sg_entry);
		c->RDMA_send_buf[subchannel][i].desc.sr.comp_type = VAPI_SIGNALED;
		c->RDMA_send_buf[subchannel][i].desc.sr.opcode = VAPI_RDMA_WRITE;
		c->RDMA_send_buf[subchannel][i].desc.sr.remote_qkey = 0;
		c->RDMA_send_buf[subchannel][i].desc.sr.remote_qp = 0; 
		c->RDMA_send_buf[subchannel][i].desc.sr.id = (aint_t) 
				&(c->RDMA_send_buf[subchannel][i]); 
		c->RDMA_send_buf[subchannel][i].desc.sr.r_key = 
				c->remote_RDMA_buf_hndl[subchannel].rkey;
		c->RDMA_send_buf[subchannel][i].desc.sr.remote_addr = 
			    (VAPI_virt_addr_t)(MT_virt_addr_t)
				(&((c->remote_RDMA_buf[subchannel] + i)->head_flag));
		c->RDMA_send_buf[subchannel][i].desc.sr.sg_lst_len = 1;

    }
	
}

#endif
static void ib_vapi_finalize()
{
   
    call_result_t ret;
    int i, j;
    
    D_PRINT("Entering ib_vapi_finalize()\n");

#ifdef MCST_SUPPORT

    if (is_mcast_enabled) {
        ud_cleanup();
    }
    
#ifdef _SMP_    
    if (smpi.my_local_id == 0){
        leave_mcgrp();
    }
#else
    leave_mcgrp();
#endif

#endif


#if defined(RDMA_FAST_PATH)

    if (viadev_num_rdma_buffer) {
    
	/* unpin rdma buffers */

#if defined(SINGLE_RDMA_BUFFER)
        ret = f defined(RDMA_FAST_PATH)VAPI_deregister_mr(
						viadev.nic, viadev.RDMA_buf_hndl.hndl);
        if (VAPI_OK != ret) {
            vapidev_error_abort(VAPI_RETURN_ERR,
                                "could unpin Fast RDMA buffer");
        }

        if (viadev.RDMA_bufs)
            free(viadev.RDMA_bufs);
        viadev.RDMA_bufs = NULL;
#else

    for(i = 0; i < viadev.np; i++){
		viadev_connection_t *c = &viadev.connections[i];

		if (i == viadev.me) { 
			continue;
		}

		for (j = 0; j < num_subchannels; j++) {

	    	VAPI_hca_hndl_t hca = viadev.nic[j/num_ports];

	    	ret = VAPI_deregister_mr(hca, c->RDMA_send_buf_hndl[j].hndl);

        	if (VAPI_OK != ret) {
        		vapidev_error_abort(VAPI_RETURN_ERR,
            	        "could unpin send rdma buffer");
        	}

		    ret = VAPI_deregister_mr(hca, c->RDMA_recv_buf_hndl[j].hndl);

    	    if (VAPI_OK != ret) {
        		vapidev_error_abort(VAPI_RETURN_ERR,
                	    "could unpin recv rdma buffer");
        	}


	    	/* free the buffers */
	    	free(c->RDMA_send_buf_orig[j]);
	    	free(c->RDMA_recv_buf_orig[j]);
		}
#endif
    }
}
#endif

    if (NULL == viadev.qp_hndl) {
    	vapidev_error_abort(GEN_EXIT_ERR, "Null queue pair handle");
    }

    for (i = 0; i < viadev.np; i++) {
		viadev_connection_t *c = &viadev.connections[i];

		if ( i == viadev.me ) {
			continue;
		}

		for (j = 0; j < num_subchannels; j++) {
			VAPI_hca_hndl_t hca = viadev.nic[j/num_ports];
			ret = VAPI_destroy_qp(hca, c->vi[j]);
    		if (VAPI_OK != ret) {
        		vapidev_error_abort(VAPI_RETURN_ERR, "could not destroy QP");
    		}

		}
    }

    free(viadev.qp_hndl);
    free(viadev.qp_prop);

#ifdef USE_MPD_RING
    ret = VAPI_destroy_qp(viadev.nic[0], viadev.boot_qp_hndl[0]);
    if (VAPI_OK != ret) {
        vapidev_error_abort(VAPI_RETURN_ERR, "could not destroy lhs QP");
    }

    ret = VAPI_destroy_qp(viadev.nic[0], viadev.boot_qp_hndl[1]);
    if (VAPI_OK != ret) {
        vapidev_error_abort(VAPI_RETURN_ERR, "could not destroy rhs QP");
    }

    ret = VAPI_destroy_cq(viadev.nic[0], viadev.boot_cq_hndl);
    if (VAPI_OK != ret) {
        vapidev_error_abort(VAPI_RETURN_ERR, "could not destroy CQ");
    }
#endif
    
 deallocate_vbufs(viadev.nic);

    for (j = 0; j < num_hcas; j++) {

		ret = VAPI_destroy_cq(viadev.nic[j],viadev.cq_hndl[j]);

    	if (VAPI_OK != ret) {
    		vapidev_error_abort(VAPI_RETURN_ERR, "could not destroy CP");
    	}


		/* unregister all user buffer registration */
		while ( dreg_evict() ) ;

		ret = VAPI_dealloc_pd(viadev.nic[j], viadev.ptag[j]);

	    if (VAPI_EBUSY == ret) {
		    D_PRINT("PD busy\n");
    	}
	    if (VAPI_EINVAL_PD_HNDL == ret) {
    		D_PRINT("PD invalid PD\n");
	    }	
    	if (VAPI_EINVAL_HCA_HNDL == ret) {
	    	D_PRINT("HCA invalid PD\n");
	    }
    	if (VAPI_EPERM == ret) {
    		D_PRINT("PD invalid permission\n");
    	}


		/* to release all resources */
		EVAPI_release_hca_hndl(viadev.nic[j]);



    }
    /* note that HCA open is system wide, 
	 * not per application. Thus, no HCA close */
}

#if defined(_SMP_) || defined(USE_MPD_BASIC) || defined(USE_MPD_RING)
/* gethostname does return a NULL-terminated string if
 * the hostname length is less than "HOSTNAME_LEN". Otherwise,
 * it is unspecified.
 */
static int get_host_id(char *myhostname, int hostname_len)
{
    int host_id = 0;
    struct hostent *hostent;

    hostent = gethostbyname(myhostname);
    if (hostent == NULL) {
        vapidev_error_abort(GEN_EXIT_ERR,
                            "Null value returned by gethostbyname");
    }

    host_id = ((struct in_addr *) hostent->h_addr_list[0])->s_addr;

    return host_id;
}

#endif

static void ib_rank_lid_table(int *alladdrs, int local_addr_len)
{
    int i, k;

#ifdef _SMP_
    int *hostnames_j = NULL;
	int j;

    SET_ORIGINAL_MALLOC_HOOKS;

    hostnames_j = (int *) malloc(viadev.np * sizeof(int));
    smpi.local_nodes = (int *) malloc(viadev.np * sizeof(int));

    if (hostnames_j == NULL || smpi.local_nodes == NULL) {
        vapidev_error_abort(GEN_EXIT_ERR,
                            "malloc: in ib_rank_lid_table for SMP");
    }

    SAVE_MALLOC_HOOKS;
    SET_MVAPICH_MALLOC_HOOKS;
#endif

#if defined(USE_MPD_BASIC) || defined (USE_MPD_RING)
    for (i = 0; i < viadev.np; i++) {

        char *alladdr_inv;
        char *temp_ptr;
        char temp_lid[12];
        char temp_qp[12];
        char temp_hostid[12];

        if (i == viadev.me)
            continue;

        alladdr_inv = (char *) alladdrs + i * 
	    ((viadev.np * num_subchannels + 1) * QPLEN_XDR);

	for (k = 0; k < num_subchannels ; k++) {

            /* Get the qp # */
            temp_ptr = alladdr_inv 
		+ (viadev.me * num_subchannels + k) * QPLEN_XDR;
            strncpy(temp_qp, temp_ptr, QPLEN_XDR);
            temp_qp[QPLEN_XDR] = '\0';
	    viadev.qp_table[i*num_subchannels+k] = atoi(temp_qp);

            /* Get hca lid # */
            temp_ptr = alladdr_inv + (i * num_subchannels + k) * QPLEN_XDR;
            strncpy(temp_lid, temp_ptr, QPLEN_XDR);
            temp_lid[QPLEN_XDR] = '\0';
	    viadev.lid_table[i*num_subchannels+k] = atoi(temp_lid);

	    D_PRINT("rank[%d] channel %d : hca_lid=%d, qp_num=%d\n",
		    i, k, 
		    viadev.lid_table[i*num_subchannels+k], 
		    viadev.qp_table[i*num_subchannels+k]); 
        }

	/* Get hostid # */
	temp_ptr = alladdr_inv + viadev.np *num_subchannels * QPLEN_XDR;
	strncpy(temp_hostid, temp_ptr, QPLEN_XDR);
	temp_hostid[QPLEN_XDR] = '\0';

#ifdef _SMP_
        hostnames_j[i] = atoi(temp_hostid);
#endif
	D_PRINT("hostid [%d]: hostid=%d\n", i, hostnames_j[i]);
    }
#else       /* (defined(USE_MPD_BASIC) || defined (USE_MPD_RING)) */

    for (i = 0; i < viadev.np; i++) {
		for(k = 0; k < num_subchannels ; k++){
	        viadev.lid_table[i* num_subchannels + k] = alladdrs[i* num_subchannels + k];
    	    viadev.qp_table[i* num_subchannels + k] = alladdrs[i* num_subchannels + k + viadev.np*num_subchannels];
        	D_PRINT("rank_lid[%d]: hca_lid=%d, qp_num=%d\n",
            	    i, viadev.lid_table[i], viadev.qp_table[i]);

		}
#ifdef _SMP_
        hostnames_j[i] = alladdrs[2 * num_subchannels * viadev.np + i];
#endif 
    }	

#endif      /* End of ! (defined(USE_MPD_BASIC) || defined (USE_MPD_RING)) */

#ifdef _SMP_

    smpi.num_local_nodes = 0;
    for (j = 0; j < viadev.np; j++) {
        if (hostnames_j[viadev.me] == hostnames_j[j]) {
            if (j == viadev.me) {
                smpi.my_local_id = smpi.num_local_nodes;
            }
            smpi.local_nodes[j] = smpi.num_local_nodes;
            smpi.l2g_rank[smpi.num_local_nodes] = j;
            smpi.num_local_nodes++;
        } else {
            smpi.local_nodes[j] = -1;
        }
    }
    free(hostnames_j);
#endif
}

int rank_to_hca_lid(int i, int j)
{
    return viadev.lid_table[i * num_subchannels + j];
}

int rank_to_qp_num(int i, int j)
{
    return viadev.qp_table[i * num_subchannels + j];
}


#ifdef USE_MPD_RING

/* A ring-based barrier: process 0 initiates two tokens
 * going clockwise and counter-clockwise, respectively.
 * The return of these two tokens completes the barrier.
 */
static int bootstrap_barrier(void)
{
    /* work entries related variables */
    VAPI_sr_desc_t sr;
    VAPI_sg_lst_entry_t sg_entry_s;

    /* completion related variables */
    VAPI_cq_hndl_t cq_hndl;
    VAPI_wc_desc_t rc;

    /* Memory related variables */
    VIP_MEM_HANDLE *mem_handle;
    int unit_length;
    char *send_addr;
    char *temp_buff;
    int barrier_recvd = 0;
    int ret;

    /* Find out the barrier address */
    unit_length = QPLEN_XDR * (viadev.np + 1);
    temp_buff = (char *) barrier_buff + viadev.me * unit_length;
    mem_handle = barrier_mem_hndl;

    sr.opcode = VAPI_SEND;
    sr.comp_type = VAPI_SIGNALED;
    sr.sg_lst_len = 1;
    sr.sg_lst_p = &sg_entry_s;
    sr.remote_qkey = 0;
    sg_entry_s.lkey = mem_handle->lkey;
    cq_hndl = viadev.boot_cq_hndl;

    if (viadev.me == 0) {
        /* send_lhs(); send_rhs(); recv_lhs(); recv_rhs(); */
        sr.id = 0;
        sg_entry_s.len = QPLEN_XDR;
        send_addr = temp_buff + unit_length;
        snprintf(send_addr, QPLEN_XDR, "000");
        sg_entry_s.addr = (VAPI_virt_addr_t) (MT_virt_addr_t) send_addr;
        ret = VAPI_post_sr(viadev.nic[0], viadev.boot_qp_hndl[1], &sr);
        if (VAPI_OK != ret) {
            vapidev_error_abort(VAPI_RETURN_ERR, "Error posting send!\n");
        }
        sr.id = 1;
        sg_entry_s.len = 2 * QPLEN_XDR;
        send_addr = temp_buff + unit_length + QPLEN_XDR;
        snprintf(send_addr, 2 * QPLEN_XDR, "111");
        sg_entry_s.addr = (VAPI_virt_addr_t) (MT_virt_addr_t) send_addr;
        ret = VAPI_post_sr(viadev.nic[0], viadev.boot_qp_hndl[0], &sr);
        if (VAPI_OK != ret) {
            vapidev_error_abort(VAPI_RETURN_ERR,
                                "Error posting second send!\n");
        }

        while (barrier_recvd < 2) {
            ret = VAPI_CQ_EMPTY;
            rc.opcode = VAPI_CQE_INVAL_OPCODE;

            /* Polling CQ */
            ret = VAPI_poll_cq(viadev.nic[0], cq_hndl, &rc);
            if (ret == VAPI_OK && rc.opcode == VAPI_CQE_RQ_SEND_DATA) {
                D_PRINT("Receive forward %s reverse %s id %d \n",
                        (char *) temp_buff,
                        (char *) (temp_buff + QPLEN_XDR), (int) rc.id);

                barrier_recvd++;
            } else if (ret == VAPI_EINVAL_HCA_HNDL
                       || ret == VAPI_EINVAL_CQ_HNDL) {
                fprintf(stderr, "[%d](%s:%d)invalid send CQ or DEV hndl\n",
                        viadev.me, __FILE__, __LINE__);
                fflush(stderr);
            }
        }                       /* end of while loop */
    } else {
        /* recv_lhs(); send_rhs(); recv_rhs(); send_lhs(); */
        while (barrier_recvd < 2) {
            ret = VAPI_CQ_EMPTY;
            rc.opcode = VAPI_CQE_INVAL_OPCODE;

            /* Polling CQ */
            ret = VAPI_poll_cq(viadev.nic[0], cq_hndl, &rc);
            if (ret == VAPI_OK && rc.opcode == VAPI_CQE_RQ_SEND_DATA) {
                D_PRINT("Received forward %s reverse %s id %d\n",
                        (char *) temp_buff,
                        (char *) (temp_buff + QPLEN_XDR), (int) rc.id);

                barrier_recvd++;
                sr.id = rc.id + 1 - viadev.np;
                sg_entry_s.len = (sr.id + 1) * QPLEN_XDR;
                send_addr = temp_buff + sr.id * QPLEN_XDR;
                sg_entry_s.addr =
                    (VAPI_virt_addr_t) (MT_virt_addr_t) send_addr;
                ret =
                    VAPI_post_sr(viadev.nic[0],
                                 viadev.boot_qp_hndl[1 - sr.id], &sr);

                if (VAPI_OK != ret) {
                    vapidev_error_abort(VAPI_RETURN_ERR,
                                        "Error posting second send!\n");
                }
            } else if (ret == VAPI_EINVAL_HCA_HNDL
                       || ret == VAPI_EINVAL_CQ_HNDL) {
                fprintf(stderr, "[%d](%s:%d)invalid CQ or DEV hndl\n",
                        viadev.me, __FILE__, __LINE__);
                fflush(stderr);
            }
        }
    }
    return (1);
}

static void ib_vapi_enable_ring(int lhs, int rhs, char *ring_addr)
{
    VAPI_qp_attr_t qp_attr;
    VAPI_qp_cap_t qp_cap;
    VAPI_qp_attr_mask_t qp_attr_mask;

    int i;
    VAPI_ret_t ret;

    /* Get lhs and rhs hca_lid and qp # */
    char temp_str[QPLEN_XDR + 1];
    char *temp_ptr;

    temp_str[QPLEN_XDR] = '\0';

    for (i = 0; i < 2; i++) {
        /* For hca_lid */
        temp_ptr = ring_addr + i * (QPLEN_XDR * 3);
        strncpy(temp_str, temp_ptr, QPLEN_XDR);
        D_PRINT("Got hca_lid %s \n", temp_str);
        viadev.boot_tb[i][0] = atoi(temp_str);

        temp_ptr += QPLEN_XDR * (2 - i);

        /* qp # to me usually on rhs unless I am at the beginning */
        strncpy(temp_str, temp_ptr, QPLEN_XDR);
        D_PRINT("Got queue_pair %s \n", temp_str);
        viadev.boot_tb[i][1] = atoi(temp_str);
    }

    /* Modifying  QP to INIT */
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);
    qp_attr.qp_state = VAPI_INIT;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QP_STATE);
    qp_attr.pkey_ix = DEFAULT_PKEY_IX;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_PKEY_IX);
    qp_attr.port = viadev_default_port;        //port..  ?? TO DO
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_PORT);
    qp_attr.remote_atomic_flags = VAPI_EN_REM_WRITE | VAPI_EN_REM_READ;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_REMOTE_ATOMIC_FLAGS);

    {
        ret = VAPI_modify_qp(viadev.nic[0], viadev.boot_qp_hndl[0], &qp_attr,
                             &qp_attr_mask, &qp_cap);
        if (VAPI_OK != ret) {
            vapidev_error_abort(VAPI_RETURN_ERR,
                                "Could not modify lhs qp to INIT!\n");
        }
        ret = VAPI_modify_qp(viadev.nic[0], viadev.boot_qp_hndl[1], &qp_attr,
                             &qp_attr_mask, &qp_cap);
        if (VAPI_OK != ret) {
            vapidev_error_abort(VAPI_RETURN_ERR,
                                "Could not modify rhs qp to INIT!\n");
        }
    }

    /**********************  INIT --> RTR  ************************/
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);
    qp_attr.qp_state = VAPI_RTR;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QP_STATE);
    qp_attr.qp_ous_rd_atom = DEFAULT_QP_OUS_RD_ATOM;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QP_OUS_RD_ATOM);
    qp_attr.path_mtu = DEFAULT_MTU;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_PATH_MTU);
    qp_attr.rq_psn = DEFAULT_PSN;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_RQ_PSN);
    qp_attr.pkey_ix = DEFAULT_PKEY_IX;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_PKEY_IX);
    qp_attr.min_rnr_timer = DEFAULT_MIN_RNR_TIMER;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_MIN_RNR_TIMER);

    qp_attr.av.sl = DEFAULT_SERVICE_LEVEL;
    qp_attr.av.grh_flag = FALSE;
    qp_attr.av.static_rate = DEFAULT_STATIC_RATE;
    qp_attr.av.src_path_bits = DEFAULT_SRC_PATH_BITS;
    //QP_ATTR_MASK_SET(qp_attr_mask,QP_ATTR_AV);

    /* lhs */
    for (i = 0; i < 2; i++) {
        qp_attr.dest_qp_num = viadev.boot_tb[i][1];
        QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_DEST_QP_NUM);
        qp_attr.av.dlid = viadev.boot_tb[i][0];
        QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_AV);

        D_PRINT("Remote LID[%d] QP=%d\n", qp_attr.av.dlid,
                qp_attr.dest_qp_num);
        ret =
            VAPI_modify_qp(viadev.nic[0], viadev.boot_qp_hndl[i], &qp_attr,
                           &qp_attr_mask, &qp_cap);
        if (VAPI_OK != ret) {
            vapidev_error_abort(VAPI_RETURN_ERR,
                                "Could not modify qp to RTR!\n");
        }
        D_PRINT("Modified to RTR..Qp[%d]\n", i);
    }

    /************** RTS *******************/
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);
    qp_attr.qp_state = VAPI_RTS;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QP_STATE);
    qp_attr.sq_psn = DEFAULT_PSN;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_SQ_PSN);
    qp_attr.timeout = DEFAULT_TIME_OUT;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_TIMEOUT);
    qp_attr.retry_count = DEFAULT_RETRY_COUNT;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_RETRY_COUNT);
    qp_attr.rnr_retry = DEFAULT_RNR_RETRY;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_RNR_RETRY);
    qp_attr.ous_dst_rd_atom = DEFAULT_MAX_RDMA_DST_OPS;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_OUS_DST_RD_ATOM);

    {
        ret = VAPI_modify_qp(viadev.nic[0], viadev.boot_qp_hndl[0],
                             &qp_attr, &qp_attr_mask, &qp_cap);
        if (VAPI_OK != ret) {
            vapidev_error_abort(VAPI_RETURN_ERR,
                                "Could not modify lhs qp to RTS!\n");
        }
        ret = VAPI_modify_qp(viadev.nic[0], viadev.boot_qp_hndl[1],
                             &qp_attr, &qp_attr_mask, &qp_cap);
        if (VAPI_OK != ret) {
            vapidev_error_abort(VAPI_RETURN_ERR,
                                "Could not modify rhs qp to RTS!\n");
        }
        D_PRINT("Modified to RTS..Qp[%d]\n", i);
    }
}

/* Using a ring of queue pairs to exchange all the queue_pairs,
 * If mpd is used,  only info about lhs and rsh are provided. */


static void ib_vapi_bootstrap_ring(int lhs, int rhs,
                                   char *ring_addr, char *local_addr,
                                   char *alladdrs,
                                   VIP_MEM_HANDLE * mem_handle)
{
    int i, j, ret;
    int recv_index;

    /* Now enable the queue pairs and post descriptors */
    ib_vapi_enable_ring(lhs, rhs, ring_addr);

    /* Register alladdrs and post receive descriptors */
    {
        /* work entries related variables */
        VAPI_rr_desc_t rr;
        VAPI_sg_lst_entry_t sg_entry_r;
        VAPI_sr_desc_t sr;
        VAPI_sg_lst_entry_t sg_entry_s;

        /* completion related variables */
        VAPI_cq_hndl_t cq_hndl;
        VAPI_wc_desc_t rc, sc;

        /* Memory related variables */
        VAPI_mrw_t mr_in, mr_out;

        int unit_length;
        char *dest_loc;
        char *recv_addr;
        char *send_addr;

        /* Same as local_addr_len */
        unit_length = (viadev.np * num_subchannels + 1) * QPLEN_XDR;

        mr_in.acl = VAPI_EN_LOCAL_WRITE | VAPI_EN_REMOTE_WRITE;
        mr_in.l_key = 0;
        mr_in.r_key = 0;
        mr_in.pd_hndl = viadev.ptag[0];
        mr_in.start = (VAPI_virt_addr_t) (virt_addr_t) alladdrs;
        mr_in.size = viadev.np * unit_length;
        mr_in.type = VAPI_MR;

        ret = VAPI_register_mr(viadev.nic[0],
                               &mr_in, &mem_handle->hndl, &mr_out);
        if (VAPI_OK != ret) {
            vapidev_error_abort(VAPI_RETURN_ERR,
                                "Error registering memory!");
        }
        mem_handle->lkey = mr_out.l_key;
        mem_handle->rkey = mr_out.r_key;
        /* Copy local_addr to the correct slot in alladdrs
         * and  post receive descriptors for all_addr */
        dest_loc = alladdrs + viadev.me * unit_length;
        strncpy(dest_loc, local_addr, unit_length);

        recv_index = 0;

        /* Post receive for all_addr */
        for (j = 0; j < viadev.np + 1; j++) {
            /* The last two entries are used for a barrier,
             * they overlap the local address */
            if ((j + 1) < viadev.np) {
                recv_addr = alladdrs + unit_length *
                    ((viadev.me + viadev.np - j - 1) % viadev.np);
            } else if (j < viadev.np) {
                recv_addr = alladdrs + unit_length * viadev.me;
            } else {
                recv_addr = alladdrs + unit_length * viadev.me + QPLEN_XDR;
            }

            /* Fillup a recv descriptor */
            rr.comp_type = VAPI_SIGNALED;
            rr.opcode = VAPI_RECEIVE;
            rr.id = j;
            rr.sg_lst_len = 1;
            rr.sg_lst_p = &(sg_entry_r);
            sg_entry_r.lkey = mem_handle->lkey;
            sg_entry_r.addr = (VAPI_virt_addr_t) (virt_addr_t) recv_addr;

            if ((j + 1) >= viadev.np)
                sg_entry_r.len = (j + 2 - viadev.np) * QPLEN_XDR;
            else
                sg_entry_r.len = unit_length;

            /* Post the recv descriptor */
            if (j < viadev.np) {
                ret =
                    VAPI_post_rr(viadev.nic[0], viadev.boot_qp_hndl[0], &rr);
            } else {
                ret =
                    VAPI_post_rr(viadev.nic[0], viadev.boot_qp_hndl[1], &rr);
            }
            if (VAPI_OK != ret) {
                vapidev_error_abort(VAPI_RETURN_ERR,
                                    "Error posting recv!");
            }
        }
        /* synchronize all the processes */
        pmgr_barrier();

        recv_index = 0;

        /* transfer all the addresses */
        for (j = 0; j < viadev.np - 1; j++) {

            send_addr = alladdrs +
                unit_length * ((viadev.me + viadev.np - j) % viadev.np);

            /* send to rhs */
            sr.opcode = VAPI_SEND;
            sr.comp_type = VAPI_SIGNALED;
            sr.id = j;
            sr.sg_lst_len = 1;
            sr.sg_lst_p = &sg_entry_s;

            sr.remote_qkey = 0;
            sg_entry_s.addr =
                (VAPI_virt_addr_t) (MT_virt_addr_t) send_addr;
            sg_entry_s.len = unit_length;
            sg_entry_s.lkey = mem_handle->lkey;

            ret = VAPI_post_sr(viadev.nic[0], viadev.boot_qp_hndl[1], &sr);
            if (VAPI_OK != ret) {
                vapidev_error_abort(VAPI_RETURN_ERR,
                                    "Error posting send!");
            }

            /* recv from lhs */
            cq_hndl = viadev.boot_cq_hndl;

            ret = VAPI_CQ_EMPTY;
            rc.opcode = VAPI_CQE_INVAL_OPCODE;
            rc.id = -1;

            do {
                ret = VAPI_poll_cq(viadev.nic[0], cq_hndl, &rc);
                if (ret == VAPI_OK && rc.opcode == VAPI_CQE_RQ_SEND_DATA) {
                    /* Detect any earlier message */
                    if (rc.id != recv_index) {
                        fprintf(stderr,
                                "[%d](%s:%d)unexpected message %d vs %d\n",
                                viadev.me, __FILE__, __LINE__,
                                rc.id, recv_index);
                        exit(1);
                    } else {
                        D_PRINT("expected message %d\n", rc.id);
                        recv_index = rc.id + 1;
                    }
                } else if (ret == VAPI_EINVAL_HCA_HNDL
                           || ret == VAPI_EINVAL_CQ_HNDL) {
                    fprintf(stderr,
                            "[%d](%s:%d)invalid CQ or DEV hndl\n",
                            viadev.me, __FILE__, __LINE__);
                    fflush(stderr);

                }
            } while (rc.opcode != VAPI_CQE_RQ_SEND_DATA);       /* do-while loop */
        }                       /* end of for loop */

    }                           /* end of alltoall exchange */
}
#endif

#ifdef MCST_SUPPORT

/* Creation of a separate UD_QP and attaching the QP to the mcst_grp */
static void ib_ud_qp_init(void)
{
    VAPI_qp_init_attr_t qp_init_attr;
    VAPI_qp_attr_t qp_attr;
    VAPI_qp_cap_t qp_cap;
    VAPI_qp_attr_mask_t qp_attr_mask;
    VAPI_ret_t ret;

    unsigned int act_num_cqe;

    ret =
        VAPI_create_cq(viadev.nic[NIC_0], DEFAULT_MAX_CQ_SIZE,
                       &viadev.ud_scq_hndl, &act_num_cqe);

    if (VAPI_OK != ret) {
        vapidev_error_abort(VAPI_RETURN_ERR,
                            "Cannot allocate CQ for UD-QP");
    }

    D_PRINT("Created UD - CQ = %d with %d entries\n",
            viadev.ud_scq_hndl, act_num_cqe);

    ret =
        VAPI_create_cq(viadev.nic[NIC_0], DEFAULT_MAX_CQ_SIZE,
                       &viadev.ud_rcq_hndl, &act_num_cqe);
    if (VAPI_OK != ret) {
        vapidev_error_abort(VAPI_RETURN_ERR,
                            "Cannot allocate CQ for UD-QP");
    }

    D_PRINT("Created UD - CQ = %d with %d entries\n",
            viadev.ud_rcq_hndl, act_num_cqe);

    qp_init_attr.cap.max_oust_wr_rq = DEFAULT_MAX_WQE;
    qp_init_attr.cap.max_oust_wr_sq = DEFAULT_MAX_WQE;
    qp_init_attr.cap.max_sg_size_rq = DEFAULT_MAX_SG_LIST;
    qp_init_attr.cap.max_sg_size_sq = DEFAULT_MAX_SG_LIST;
    qp_init_attr.pd_hndl = viadev.ptag[NIC_0];
    qp_init_attr.rdd_hndl = 0;
    qp_init_attr.rq_cq_hndl = viadev.ud_rcq_hndl;
    qp_init_attr.sq_cq_hndl = viadev.ud_scq_hndl;
    qp_init_attr.rq_sig_type = VAPI_SIGNAL_REQ_WR;
    qp_init_attr.sq_sig_type = VAPI_SIGNAL_REQ_WR;
    qp_init_attr.ts_type = VAPI_TS_UD;
    ret = VAPI_create_qp(viadev.nic[NIC_0], &qp_init_attr,
                         &viadev.ud_qp_hndl, &viadev.ud_qp_prop);
    if (VAPI_OK != ret) {
        vapidev_error_abort(VAPI_RETURN_ERR, "Cannot create UD-QP");
    }
    D_PRINT(" UD QP[%d] created ..hndl=%d\n",
            viadev.ud_qp_prop.qp_num, viadev.ud_qp_hndl);

    /* Modifying  QP to INIT */
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);
    qp_attr.qp_state = VAPI_INIT;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QP_STATE);
    qp_attr.pkey_ix = DEFAULT_PKEY_IX;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_PKEY_IX);
    qp_attr.port = viadev_default_port;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_PORT);
    qp_attr.qkey = 0;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QKEY);

    ret = VAPI_modify_qp(viadev.nic[NIC_0], viadev.ud_qp_hndl, &qp_attr,
                         &qp_attr_mask, &qp_cap);
    if (VAPI_OK != ret) {
        vapidev_error_abort(VAPI_RETURN_ERR,
                            "Could not modify UD-QP to INIT");
    }
    D_PRINT("Modified UD to init..Qp\n");

    /**********************  INIT --> RTR  ************************/
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);
    qp_attr.qp_state = VAPI_RTR;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QP_STATE);

    ret =
        VAPI_modify_qp(viadev.nic[NIC_0], viadev.ud_qp_hndl, &qp_attr,
                       &qp_attr_mask, &qp_cap);
    if (VAPI_OK != ret) {
        vapidev_error_abort(VAPI_RETURN_ERR,
                            "Could not modify UD-QP to RTR");
    }
    D_PRINT("Modified UD to RTR..Qp\n");


    /********************* RTR --> RTS  * *******************/
    QP_ATTR_MASK_CLR_ALL(qp_attr_mask);
    qp_attr.qp_state = VAPI_RTS;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_QP_STATE);
    qp_attr.sq_psn = DEFAULT_PSN;
    QP_ATTR_MASK_SET(qp_attr_mask, QP_ATTR_SQ_PSN);
    ret =
        VAPI_modify_qp(viadev.nic[NIC_0], viadev.ud_qp_hndl, &qp_attr,
                       &qp_attr_mask, &qp_cap);
    if (VAPI_OK != ret) {
        vapidev_error_abort(VAPI_RETURN_ERR,
                            "Could not modify UD-QP to RTS");
    }
    D_PRINT("Modified UD to RTS..Qp\n");

    /* Attach to the mulitcast grp - currently all nodes are in bcast grp */

    ret =
        VAPI_attach_to_multicast(viadev.nic[NIC_0], viadev.mcg_dgid,
                                 viadev.ud_qp_hndl, 0);
    if (VAPI_OK != ret) {
        vapidev_error_abort(VAPI_RETURN_ERR,
                            "Could not attach UD-QP to MCGrp\n");
    }
}

static void ud_mcst_av_init(void)
{

    VAPI_ret_t ret;


    viadev.av.dgid[0] = 0;
#ifdef MINISM
    viadev.av.dlid = 0xC000;
#endif

#if defined(OSM_VENDOR_INTF_TS) || defined(OSM_VENDOR_INTF_MTL)
    viadev.av.dlid = mlid;
#endif

    viadev.av.grh_flag = TRUE;
    viadev.av.flow_label = DEFAULT_FLOW_LABEL;
    viadev.av.hop_limit = DEFAULT_HOP_LIMIT;
    viadev.av.sgid_index = 0;
    memcpy(viadev.av.dgid, viadev.mcg_dgid, sizeof(IB_gid_t));
    viadev.av.sl = DEFAULT_SERVICE_LEVEL;
    viadev.av.port = viadev_default_port;
    viadev.av.src_path_bits = DEFAULT_SRC_PATH_BITS;
    viadev.av.static_rate = DEFAULT_STATIC_RATE;
    viadev.av.traffic_class = DEFAULT_TRAFFIC_CLASS;

    ret = VAPI_create_addr_hndl(viadev.nic[NIC_0], viadev.ptag[NIC_0],
                                &viadev.av, &viadev.av_hndl);

    if (VAPI_OK != ret) {
        vapidev_error_abort(VAPI_RETURN_ERR,
                            "Could not create address vector\n");
    }
}

static void alloc_ack_region()
{

    VAPI_ret_t result;
    VAPI_mrw_t mr_in, mr_out;
    int mem_total;
    int alignment = 4;          /* aligning to a 4 byte boundary */

    if (LEFT_OVERS == 0){
    mem_total =
        (NUM_CHILD * NUM_COROOTS + NUM_COROOTS + NUM_WBUFFERS) * ACK_LEN
        + alignment - 1;
    }
    else{
        if (viadev.me == 0){
            mem_total = ((NUM_CHILD + LEFT_OVERS) * NUM_COROOTS + NUM_COROOTS +
                 NUM_WBUFFERS) * ACK_LEN
                + alignment - 1;
        }
        else {
            if (viadev.me >= NUM_CHILD * NUM_COROOTS){
                mem_total = (viadev.np  + 1 + NUM_WBUFFERS) * ACK_LEN
                    + alignment - 1;
            }
            else{
                mem_total = (NUM_CHILD * NUM_COROOTS + NUM_COROOTS +
                        NUM_WBUFFERS) * ACK_LEN + alignment - 1;
            }
        }

    }
        
    
    viadev.bcast_info.ack_region = malloc(mem_total);
    memset(viadev.bcast_info.ack_region, 0, mem_total);

    viadev.bcast_info.ack_buffer =
        (void
         *) (((aint_t) viadev.bcast_info.ack_region +
              (aint_t) (alignment - 1))
             & ~((aint_t) alignment - 1));

    co_print("local mem:%p\n", viadev.bcast_info.ack_buffer);
    mr_in.acl = VAPI_EN_LOCAL_WRITE | VAPI_EN_REMOTE_WRITE;
    mr_in.l_key = 0;
    mr_in.r_key = 0;
    mr_in.pd_hndl = viadev.ptag[NIC_0];
    mr_in.size = mem_total - alignment - 1;
    mr_in.start =
        (VAPI_virt_addr_t) (virt_addr_t) (viadev.bcast_info.ack_buffer);
    mr_in.type = VAPI_MR;

    result = VAPI_register_mr(viadev.nic[NIC_0], &mr_in,
                              &(viadev.bcast_info.ack_mem_hndl.hndl),
                              &mr_out);
    if (VAPI_OK != result) {
        vapidev_error_abort(VAPI_RETURN_ERR,
                            "Could not register memory\n");
    }
    viadev.bcast_info.ack_mem_hndl.lkey = mr_out.l_key;
    viadev.bcast_info.ack_mem_hndl.rkey = mr_out.r_key;
    co_print("local r_key:%d\n", mr_out.r_key);
}

static void alloc_bcast_buffers(void)
{
    int i, j;
    int *ack_ptr = (int *) bcast_info->ack_buffer;
    int num_coroots, num_child;

    if (LEFT_OVERS == 0){
        num_coroots = NUM_COROOTS;
        num_child = NUM_CHILD;
    }
    else{
        if (viadev.me > LAST_NODE){
            num_coroots = 1;num_child = viadev.np;
        }
        else {
            if (viadev.me == 0){
                num_coroots = NUM_COROOTS;
                num_child = NUM_CHILD + LEFT_OVERS;
            }
            else {
               num_coroots = NUM_COROOTS; num_child = NUM_CHILD;
            }
        }
    }

    bcast_info->ack_cnt = malloc((sizeof(int)) * (viadev.np));
    bcast_info->msg_cnt = malloc((sizeof(int)) * (viadev.np));
    bcast_info->time_out =
        (double **) malloc((sizeof(double *)) * num_coroots);
    bcast_info->win_head = malloc((sizeof(int)) * num_coroots);
    bcast_info->win_tail = malloc((sizeof(int)) * num_coroots);
    bcast_info->buf_head = malloc((sizeof(int)) * num_coroots);
    bcast_info->buf_tail = malloc((sizeof(int)) * num_coroots);
    bcast_info->min_ack = malloc((sizeof(int)) * num_coroots);
    bcast_info->is_full = malloc((sizeof(int)) * num_coroots);
    bcast_info->co_buf = malloc((sizeof(vbuf *)) * num_coroots);


    for (i = 0; i < num_coroots; i++) {
        bcast_info->win_head[i] = bcast_info->win_tail[i] = 0;
        bcast_info->buf_head[i] = bcast_info->buf_tail[i] = 0;
        bcast_info->sbuf_head = bcast_info->sbuf_tail = 0;
        bcast_info->ack_full = 0;
        bcast_info->is_full[i] = 0;
        bcast_info->co_buf[i] = malloc((sizeof(vbuf)) * SENDER_WINDOW);
        bcast_info->time_out[i] =
            (double *) malloc((sizeof(double)) * SENDER_WINDOW);
        bcast_info->min_ack[i] = -1;

    }

    for (i = 0; i < viadev.np; i++)
        bcast_info->Bcnt[i] = 0;

    for (i = 0; i < num_child * num_coroots + num_coroots + NUM_WBUFFERS;
         i++)
        ack_ptr[i] = -1;

    for (i = 0; i < num_coroots; i++) {
        for (j = 0; j < num_child; j++)
            bcast_info->time_out[i][j] = -1;
    }

    bcast_info->bcast_called = 0;
}
void ud_cleanup(void)
{

    VAPI_ret_t ret;
    int i;
    int num_coroots;

    if (LEFT_OVERS == 0){
        num_coroots = NUM_COROOTS;
    }
    else{
        if (viadev.me > LAST_NODE){
            num_coroots = 1;
        }
        else {
            num_coroots = NUM_COROOTS;
        }
    }


    free(bcast_info->ack_cnt);
    free(bcast_info->msg_cnt);
    free(bcast_info->win_head);
    free(bcast_info->win_tail);
    free(bcast_info->buf_head);
    free(bcast_info->buf_tail);
    free(bcast_info->is_full);
    free(bcast_info->buf_credits);
    free(bcast_info->buf_updates);

    for (i = 0; i < num_coroots; i++) {
        free(bcast_info->co_buf[i]);
        free(bcast_info->time_out[i]);
    }

    free(bcast_info->time_out);
    free(bcast_info->co_buf);
    ret = VAPI_detach_from_multicast(viadev.nic[NIC_0], viadev.mcg_dgid,
                                     viadev.ud_qp_hndl, 0);


    ret = VAPI_destroy_qp(viadev.nic[NIC_0], viadev.ud_qp_hndl);

    ret = VAPI_destroy_addr_hndl(viadev.nic[NIC_0], viadev.av_hndl);
    ret = VAPI_destroy_cq(viadev.nic[NIC_0], viadev.ud_scq_hndl);
    ret = VAPI_destroy_cq(viadev.nic[NIC_0], viadev.ud_rcq_hndl);
    ret = VAPI_deregister_mr(viadev.nic[NIC_0], bcast_info->ack_mem_hndl.hndl);
    free(bcast_info->ack_buffer);


}

#endif


#ifdef PENTIUM_MHZ
double get_us(void)
{
   unsigned long long now;
   __asm __volatile (".byte 0x0f, 0x31" : "=A" (now));
   return (double)((double)now / (PENTIUM_MHZ));
}
#else
double get_us(void)
{
   struct timeval t;
   gettimeofday(&t,NULL);
   return (double)t.tv_sec * (double)1e6 + (double)t.tv_usec;
}
#endif

#ifdef MAC_OSX
void mvapich_malloc_init()
{
  /* This dylib is Mac OS X specific. For other unix platforms
     replace libc.dylib with libc.so */

  if(vt_malloc_ptr != NULL) return;
  vt_dylib = dlopen("/usr/lib/libc.dylib", RTLD_LAZY);

  if (vt_dylib == NULL)
    {
      printf("Open error: Unable to open dynamic library ...\n");
      exit(1);
    }
#if defined(VIADEV_RPUT_SUPPORT) && defined(LAZY_MEM_UNREGISTER)

  vt_malloc_ptr = dlsym(vt_dylib, "malloc");

  if (vt_malloc_ptr == NULL)
    {
      printf("Resolution error: Unable to locate symbol for malloc ...\n");
      exit(1);
    }

  vt_free_ptr = dlsym(vt_dylib, "free");

  if (vt_free_ptr == NULL)
    {
      printf("Resolution error: Unable to locate symbol for free ...\n");
      exit(1);
    }
#endif
}
#endif

